//
// ISR.cpp:
//	Implementation of class segaISR.
//
// Purpose:
//	Encapsulates functions that set up the interrupt
//	service routine and the functions actually
//	called by the ISR.
//
// Author(s): Mark Pflaging   C-serve  70233,1552
//	      Dave Stampe     Internet dstamp@watserv1.uwaterloo.ca
//
// Also see authorships in OOPGLOVE.DOC.
// Date:    5/15/92 - 9/21/92
//
#include "isr.hpp"
#include <dos.h>
#include <bios.h>
#include <stdio.h>
#include <conio.h>
#include <signal.h>

/******************* TIMER INTERRUPT STUFF ****************/

#define TIMER_VECT 8		// timer chip int. vector
#define UNUSED_VECT 128         // user vector: used to link to std ISR

#define PIC_CMD 0x20            // constants for PIC EOI command
#define NONSPEC_EOI 0x20
#define TIMER_MODE 0x34         // timer setup command
#define TIMER_CONTROL 0x43	// timer control register port
#define TIMER_0 0x40		// timer 0 port

#define LATCH_COUNT 0x00	// command to latch count

/**********************************************************/

// how many frames between resynchronization
#define SYNC_INTERVAL 5

// Time (microseconds) of advance in vert. interrupt.  Will act as
// "vertical hold" adjustment.
#define ADVANCE 100

extern "C" void far vsync();	// pause till vert. retrace

unsigned int segaISR::clock_rate = 0;
unsigned int segaISR::clock_rate_hi = 0;
unsigned int segaISR::clock_rate_lo = 0;
unsigned int segaISR::divisor = 0;
int segaISR::syncount = SYNC_INTERVAL;
long segaISR::int_timer = 0;
int segaISR::glovecount = 0;
int segaISR::glove_rate = 2;
void interrupt (*segaISR::oldint)(...);
void (* segaISR::frame_interrupt_routine)();
void (* segaISR::glove_interrupt_routine)();

// segaISR::findSegaSpeed()
//	Routine to compute vertical frame time.
//	Call after setting video mode.
int segaISR::findSegaSpeed()
{
	vsync();
	outportb(TIMER_CONTROL, TIMER_MODE);
	outportb(TIMER_0, 0);
	// timer must count modulo 65536
	outportb(TIMER_0, 0);

	// time of vert. retrace
	disable();
	vsync();
	outportb(TIMER_CONTROL,LATCH_COUNT);
	enable();
	unsigned int oldRate = inportb(TIMER_0) & 0xFF;
	oldRate |= (inportb(TIMER_0) << 8);

	// time 2 vert. retraces later
	vsync();
	disable();
	vsync();
	outportb(TIMER_CONTROL,LATCH_COUNT);
	enable();
	unsigned int newRate = inportb(TIMER_0) & 0xFF;
	newRate |= (inportb(TIMER_0) << 8);
	// compute interrupt rate
	return ((oldRate - newRate) >> 1) - ADVANCE;
}

void segaISR::handleGlove()
{
	unsigned int olddiv = divisor;
	if (glove_interrupt_routine != NULL) glove_interrupt_routine();

	// dec divide count
	divisor += clock_rate;
	// simulate 18.2 Hz ISR if time
	if (divisor < olddiv)
		geninterrupt(UNUSED_VECT);

	outportb(PIC_CMD, NONSPEC_EOI);
	enable();
	return;
}

void interrupt segaISR::fastTimer(...)
{

	// vertical processing needed?
	disable();

	if (frame_interrupt_routine == NULL) {
		handleGlove();
		return;
	}
	if (--glovecount <= 0) glovecount = glove_rate;
	else {
		handleGlove();
		return;
	}

	int_timer++;

	syncount--;
	if (syncount < 0) {
		syncount = SYNC_INTERVAL;
		// stop timer
		outportb(TIMER_CONTROL, TIMER_MODE);
		outportb(TIMER_0, 255);
		outportb(TIMER_0, 255);
		vsync();
		// restart timer
		outportb(TIMER_CONTROL, TIMER_MODE);
		outportb(TIMER_0, clock_rate_lo);
		outportb(TIMER_0, clock_rate_hi);
	}

	if (frame_interrupt_routine != NULL) frame_interrupt_routine();
	handleGlove();
}

// SET UP FAST TIMER
void segaISR::initTimer(int speed)
{
	// set traps for error, ctrl c,
	atexit(segaISR::resetTimer);
	signal(SIGABRT, (void (*)(int))segaISR::ResetTimer);
	signal(SIGFPE,  (void (*)(int))segaISR::ResetTimer);
	signal(SIGINT,  (void (*)(int))segaISR::ResetTimer);

	// save speed for future work
	clock_rate = speed;
	clock_rate_hi = clock_rate>>8;
	clock_rate_lo = clock_rate&255;

	// set up timers
	divisor = 0;
	syncount = SYNC_INTERVAL;
	int_timer = 0;

	disable();
	if (getvect(UNUSED_VECT) == 0) {
		// setup int's if required(first run)
		oldint = getvect(TIMER_VECT);	// setup ISR vectors
		setvect(UNUSED_VECT,oldint);
		setvect(TIMER_VECT, fastTimer);
	}
	outportb(TIMER_CONTROL, TIMER_MODE);	// load timer
	outportb(TIMER_0, clock_rate_lo);
	outportb(TIMER_0, clock_rate_hi);
	enable();
}

// RESET PC TO NORMAL
#pragma argsused
void segaISR::ResetTimer(int * reglist)
{
	disable();
	setvect(TIMER_VECT, oldint);		// reset vector
	outportb(TIMER_CONTROL, TIMER_MODE);	// reset timer
	outportb(TIMER_0, 0);
	outportb(TIMER_0, 0);
	setvect(UNUSED_VECT,NULL);		// disconnect flag
	enable();

	signal(SIGINT, SIG_DFL);
	signal(SIGABRT, SIG_DFL);
	signal(SIGFPE, SIG_DFL);
}

// 1.19*x = 1000;  x = 1000/1.19 = 840
#define MILLISEC   840

// segaISR::segaISR(): SET EVERYTHING UP (init. glove, Sega first)
// Input:
//	sega_switcher: NULL if no sega/video
//	glove_handler: NULL if no glove driver
//	glove_tc: 1.19*uS (test interval) - 6500 sugg., 0 if none
//
segaISR::segaISR(void(* glove_handler)(), void(* sega_switcher)())
{
	glove_interrupt_routine = glove_handler;
	frame_interrupt_routine = sega_switcher;

	if (sega_switcher == NULL)
		// glove only: use suggested speed
		i_rate = MILLISEC;
	else {
		// use a multiple of video rate
		i_rate = findSegaSpeed();
		if (glove_handler) {
			// glove_rate is about 2x for 72 fps,
			// 3x for 60
			glove_rate = i_rate/MILLISEC;
			// handle no glove driver
			if (glove_rate < 1) glove_rate = 1;
			// come on now...
			if (glove_rate > 20) glove_rate = 20;
			i_rate /= glove_rate;
		}
	}
}
